#cython: language_level=3
# Copyright 2005-2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# See www.openfst.org for extensive documentation on this weighted
# finite-state transducer library.


from libcpp cimport bool
from libcpp.memory cimport shared_ptr
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
from libcpp.utility cimport pair
from libcpp.vector cimport vector

from cintegral_types cimport int32
from cintegral_types cimport int64
from cintegral_types cimport uint8
from cintegral_types cimport uint64
from cios cimport ostream
from cios cimport stringstream

cimport cpywrapfst as fst


# Exportable helper functions.


cdef string tostring(data) except *

cdef string weight_tostring(data) except *

cdef string path_tostring(data) except *

cdef fst.ComposeFilter _get_compose_filter(
    const string &compose_filter) except *

cdef fst.DeterminizeType _get_determinize_type(const string &det_type) except *

cdef fst.QueueType _get_queue_type(const string &queue_type) except *

cdef fst.RandArcSelection _get_rand_arc_selection(
    const string &replace_label_type) except *

cdef fst.ReplaceLabelType _get_replace_label_type(
    const string &replace_label_type,
    bool epsilon_on_replace) except *


# Weight.


cdef fst.WeightClass _get_WeightClass_or_one(const string &weight_type,
                                             weight_string) except *

cdef fst.WeightClass _get_WeightClass_or_zero(const string &weight_type,
                                              weight_string) except *


cdef class Weight:

  cdef unique_ptr[fst.WeightClass] _weight

  cdef void _check_weight(self) except *

  cpdef Weight copy(self)

  cpdef string to_string(self)

  cpdef string type(self)

  cpdef bool member(self)


cdef Weight _zero(weight_type)

cdef Weight _one(weight_type)

cdef Weight _no_weight(weight_type)

cdef Weight _plus(Weight lhs, Weight rhs)

cdef Weight _times(Weight lhs, Weight rhs)

cdef Weight _divide(Weight lhs, Weight rhs)

cdef Weight _power(Weight lhs, size_t n)


# SymbolTable.

ctypedef fst.SymbolTable * SymbolTable_ptr
ctypedef const fst.SymbolTable * const_SymbolTable_ptr


cdef class SymbolTableView:

  cdef const fst.SymbolTable *_raw(self)

  cdef void _raise_nonexistent(self) except *

  cdef const fst.SymbolTable *_raw_ptr_or_raise(self) except *

  cpdef int64 available_key(self) except *

  cpdef bytes checksum(self)

  cpdef SymbolTable copy(self)

  cpdef int64 get_nth_key(self, ssize_t pos) except *

  cpdef bytes labeled_checksum(self)

  cpdef bool member(self, key) except *

  cpdef string name(self) except *

  cpdef size_t num_symbols(self) except *

  cpdef void write(self, source) except *

  cpdef void write_text(self, source) except *

  cpdef bytes write_to_string(self)


cdef class _EncodeMapperSymbolTableView(SymbolTableView):

  # Indicates whether this view is of an input or output SymbolTable
  cdef bool _input_side

  cdef shared_ptr[fst.EncodeMapperClass] _mapper


cdef class _FstSymbolTableView(SymbolTableView):

  # Indicates whether this view is of an input or output SymbolTable
  cdef bool _input_side

  cdef shared_ptr[fst.FstClass] _fst


cdef class _MutableSymbolTable(SymbolTableView):

  cdef fst.SymbolTable *_mutable_raw(self)

  cdef fst.SymbolTable *_mutable_raw_ptr_or_raise(self) except *

  cpdef int64 add_symbol(self, symbol, int64 key=?) except *

  cpdef void add_table(self, SymbolTableView syms) except *

  cpdef void set_name(self, new_name) except *


cdef class _MutableFstSymbolTableView(_MutableSymbolTable):

  # Indicates whether this view is of an input or output SymbolTable
  cdef bool _input_side

  cdef shared_ptr[fst.MutableFstClass] _mfst


cdef class SymbolTable(_MutableSymbolTable):

  cdef unique_ptr[fst.SymbolTable] _smart_table


cdef _EncodeMapperSymbolTableView _init_EncodeMapperSymbolTableView(
    shared_ptr[fst.EncodeMapperClass] encoder, bool input_side)


cdef _FstSymbolTableView _init_FstSymbolTableView(shared_ptr[fst.FstClass] ifst,
                                                  bool input_side)


cdef _MutableFstSymbolTableView _init_MutableFstSymbolTableView(
    shared_ptr[fst.MutableFstClass] ifst, bool input_side)


cdef SymbolTable _init_SymbolTable(unique_ptr[fst.SymbolTable] table)


cpdef SymbolTable _read_SymbolTable_from_string(string state)


cdef class _SymbolTableIterator:

  cdef SymbolTableView _table
  cdef unique_ptr[fst.SymbolTableIterator] _siter


# EncodeMapper.


ctypedef fst.EncodeMapperClass * EncodeMapperClass_ptr


cdef class EncodeMapper:

  cdef shared_ptr[fst.EncodeMapperClass] _mapper

  cpdef string arc_type(self)

  cpdef string weight_type(self)

  cpdef uint8 flags(self)

  cpdef void write(self, source) except *

  cpdef bytes write_to_string(self)

  cpdef _EncodeMapperSymbolTableView input_symbols(self)

  cpdef _EncodeMapperSymbolTableView output_symbols(self)

  cdef void _set_input_symbols(self, SymbolTableView syms) except *

  cdef void _set_output_symbols(self, SymbolTableView syms) except *


cdef EncodeMapper _init_EncodeMapper(EncodeMapperClass_ptr mapper)

cpdef EncodeMapper _read_EncodeMapper_from_string(string state)


# Fst.


ctypedef fst.FstClass * FstClass_ptr
ctypedef const fst.FstClass * const_FstClass_ptr
ctypedef fst.MutableFstClass * MutableFstClass_ptr
ctypedef fst.VectorFstClass * VectorFstClass_ptr


cdef class Fst:

  cdef shared_ptr[fst.FstClass] _fst

  @staticmethod
  cdef string _local_render_svg(const string &)

  cpdef string arc_type(self)

  cpdef _ArcIterator arcs(self, int64 state)

  cpdef Fst copy(self)

  cpdef void draw(self,
                  source,
                  SymbolTableView isymbols=?,
                  SymbolTableView osymbols=?,
                  SymbolTableView ssymbols=?,
                  bool acceptor=?,
                  title=?,
                  double width=?,
                  double height=?,
                  bool portrait=?,
                  bool vertical=?,
                  double ranksep=?,
                  double nodesep=?,
                  int32 fontsize=?,
                  int32 precision=?,
                  float_format=?,
                  bool show_weight_one=?) except *

  cpdef Weight final(self, int64 state)

  cpdef string fst_type(self)

  cpdef _FstSymbolTableView input_symbols(self)

  cpdef size_t num_arcs(self, int64 state) except *

  cpdef size_t num_input_epsilons(self, int64 state) except *

  cpdef size_t num_output_epsilons(self, int64 state) except *

  cpdef _FstSymbolTableView output_symbols(self)

  cpdef string print(self,
                    SymbolTableView isymbols=?,
                    SymbolTableView osymbols=?,
                    SymbolTableView ssymbols=?,
                    bool acceptor=?,
                    bool show_weight_one=?,
                    missing_sym=?) except *

  cpdef int64 start(self)

  cpdef _StateIterator states(self)

  cpdef bool verify(self)

  cpdef string weight_type(self)

  cpdef void write(self, source) except *

  cpdef bytes write_to_string(self)


cdef class MutableFst(Fst):

  cdef shared_ptr[fst.MutableFstClass] _mfst

  cdef void _check_mutating_imethod(self) except *

  cdef void _add_arc(self, int64 state, Arc arc) except *

  cpdef int64 add_state(self)

  cpdef void add_states(self, size_t)

  cdef void _arcsort(self, sort_type=?) except *

  cdef void _closure(self, bool closure_plus=?)

  cdef void _concat(self, Fst fst2) except *

  cdef void _connect(self)

  cdef void _decode(self, EncodeMapper) except *

  cdef void _delete_arcs(self, int64 state, size_t n=?) except *

  cdef void _delete_states(self, states=?) except *

  cdef void _encode(self, EncodeMapper) except *

  cdef void _invert(self)

  cdef void _minimize(self, float delta=?, bool allow_nondet=?) except *

  cpdef _MutableArcIterator mutable_arcs(self, int64 state)

  cpdef int64 num_states(self)

  cdef void _project(self, project_type) except *

  cdef void _prune(self, float delta=?, int64 nstate=?, weight=?) except *

  cdef void _push(self,
                  float delta=?,
                  bool remove_total_weight=?,
                  bool to_final=?)

  cdef void _relabel_pairs(self, ipairs=?, opairs=?) except *

  cdef void _relabel_tables(self,
                            SymbolTableView old_isymbols=?,
                            SymbolTableView new_isymbols=?,
                            unknown_isymbol=?,
                            bool attach_new_isymbols=?,
                            SymbolTableView old_osymbols=?,
                            SymbolTableView new_osymbols=?,
                            unknown_osymbol=?,
                            bool attach_new_osymbols=?) except *

  cdef void _reserve_arcs(self, int64 state, size_t n) except *

  cdef void _reserve_states(self, int64 n)

  cdef void _reweight(self, potentials, bool to_final=?) except *

  cdef void _rmepsilon(self,
                       queue_type=?,
                       bool connect=?,
                       weight=?,
                       int64 nstate=?,
                       float delta=?) except *

  cdef void _set_final(self, int64 state, weight=?) except *

  cdef void _set_start(self, int64 state) except *

  cdef void _set_input_symbols(self, SymbolTableView syms) except *

  cdef void _set_output_symbols(self, SymbolTableView syms) except *

  cdef void _topsort(self)


cdef class VectorFst(MutableFst):

    pass


# Construction helpers.


cdef Fst _init_Fst(FstClass_ptr tfst)

cdef MutableFst _init_MutableFst(MutableFstClass_ptr tfst)

cdef Fst _init_XFst(FstClass_ptr tfst)

cpdef Fst _read_Fst(source)

cpdef Fst _read_Fst_from_string(string state)


# Iterators.


cdef class Arc:

  cdef unique_ptr[fst.ArcClass] _arc

  cpdef Arc copy(self)


cdef Arc _init_Arc(const fst.ArcClass &arc)


cdef class _ArcIterator:

  cdef shared_ptr[fst.FstClass] _fst
  cdef unique_ptr[fst.ArcIteratorClass] _aiter

  cpdef bool done(self)

  cpdef uint8 flags(self)

  cpdef void next(self)

  cpdef size_t position(self)

  cpdef void reset(self)

  cpdef void seek(self, size_t a)

  cpdef void set_flags(self, uint8 flags, uint8 mask)

  cpdef object value(self)


cdef class _MutableArcIterator:

  cdef shared_ptr[fst.MutableFstClass] _mfst
  cdef unique_ptr[fst.MutableArcIteratorClass] _aiter

  cpdef bool done(self)

  cpdef uint8 flags(self)

  cpdef void next(self)

  cpdef size_t position(self)

  cpdef void reset(self)

  cpdef void seek(self, size_t a)

  cpdef void set_flags(self, uint8 flags, uint8 mask)

  cpdef void set_value(self, Arc arc)

  cpdef object value(self)


cdef class _StateIterator:

  cdef shared_ptr[fst.FstClass] _fst
  cdef unique_ptr[fst.StateIteratorClass] _siter

  cpdef bool done(self)

  cpdef void next(self)

  cpdef void reset(self)

  cpdef int64 value(self)


# Constructive operations on Fst.


cdef Fst _map(Fst ifst, float delta=?, map_type=?, double power=?, weight=?)

cpdef Fst arcmap(Fst ifst, float delta=?, map_type=?, double power=?, weight=?)

cpdef MutableFst compose(Fst ifst1,
                         Fst ifst2,
                         compose_filter=?,
                         bool connect=?)

cpdef Fst convert(Fst ifst, fst_type=?)

cpdef MutableFst determinize(Fst ifst,
                             float delta=?,
                             det_type=?,
                             int64 nstate=?,
                             int64 subsequential_label=?,
                             weight=?,
                             bool increment_subsequential_label=?)

cpdef MutableFst difference(Fst ifst1,
                            Fst ifst2,
                            compose_filter=?,
                            bool connect=?)

cpdef MutableFst disambiguate(Fst ifst,
                              float delta=?,
                              int64 nstate=?,
                              int64 subsequential_label=?,
                              weight=?)

cpdef MutableFst epsnormalize(Fst ifst, bool eps_norm_output=?)

cpdef bool equal(Fst ifst1, Fst ifst2, float delta=?)

cpdef bool equivalent(Fst ifst1, Fst ifst2, float delta=?) except *

cpdef MutableFst intersect(Fst ifst1,
                           Fst ifst2,
                           compose_filter=?,
                           bool connect=?)

cpdef bool isomorphic(Fst ifst1, Fst ifst2, float delta=?)

cpdef MutableFst prune(Fst ifst,
                       float delta=?,
                       int64 nstate=?,
                       weight=?)

cpdef MutableFst push(Fst ifst,
                      float delta=?,
                      bool push_weights=?,
                      bool push_labels=?,
                      bool remove_common_affix=?,
                      bool remove_total_weight=?,
                      bool to_final=?)

cpdef bool randequivalent(Fst ifst1,
                          Fst ifst2,
                          int32 npath=?,
                          float delta=?,
                          select=?,
                          int32 max_length=?,
                          uint64 seed=?) except *

cpdef MutableFst randgen(Fst ifst,
                         int32 npath=?,
                         select=?,
                         int32 max_length=?,
                         bool remove_total_weight=?,
                         bool weighted=?,
                         uint64 seed=?)

cpdef MutableFst replace(pairs,
                         call_arc_labeling=?,
                         return_arc_labeling=?,
                         bool epsilon_on_replace=?,
                         int64 return_label=?)

cpdef MutableFst reverse(Fst ifst, bool require_superinitial=?)

cdef void _shortestdistance(Fst ifst,
                            vector[fst.WeightClass] *,
                            float delta=?,
                            int64 nstate=?,
                            queue_type=?,
                            bool reverse=?) except *

cpdef MutableFst shortestpath(Fst ifst,
                              float delta=?,
                              int32 nshortest=?,
                              int64 nstate=?,
                              queue_type=?,
                              bool unique=?,
                              weight=?)

cpdef Fst statemap(Fst ifst, map_type)

cpdef MutableFst synchronize(Fst ifst)


# Compiler.


cdef class Compiler:

  cdef unique_ptr[stringstream] _sstrm
  cdef string _fst_type
  cdef string _arc_type
  cdef const fst.SymbolTable *_isymbols
  cdef const fst.SymbolTable *_osymbols
  cdef const fst.SymbolTable *_ssymbols
  cdef bool _acceptor
  cdef bool _keep_isymbols
  cdef bool _keep_osymbols
  cdef bool _keep_state_numbering
  cdef bool _allow_negative_labels

  cpdef Fst compile(self)

  cpdef void write(self, expression)


# FarReader.

cdef class FarReader:

  cdef unique_ptr[fst.FarReaderClass] _reader

  cpdef string arc_type(self)

  cpdef bool done(self)

  cpdef bool error(self)

  cpdef string far_type(self)

  cpdef bool find(self, key)

  cpdef Fst get_fst(self)

  cpdef string get_key(self)

  cpdef void next(self)

  cpdef void reset(self)


# FarWriter.

cdef class FarWriter:

  cdef unique_ptr[fst.FarWriterClass] _writer

  cpdef string arc_type(self)

  cdef void close(self)

  cpdef void add(self, key, Fst ifst) except *

  cpdef bool error(self)

  cpdef string far_type(self)



#cython: language_level=3
# Copyright 2005-2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# See www.openfst.org for extensive documentation on this weighted
# finite-state transducer library.


from libcpp cimport bool
from libcpp.memory cimport unique_ptr
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.utility cimport pair

from cios cimport *
from cintegral_types cimport *


cdef extern from "<fst/util.h>" nogil:

  bool FST_FLAGS_fst_error_fatal


cdef extern from "<fst/fstlib.h>" namespace "fst" nogil:

  # FST properties.
  const uint64 kExpanded
  const uint64 kMutable
  const uint64 kError
  const uint64 kAcceptor
  const uint64 kNotAcceptor
  const uint64 kIDeterministic
  const uint64 kNonIDeterministic
  const uint64 kODeterministic
  const uint64 kNonODeterministic
  const uint64 kEpsilons
  const uint64 kNoEpsilons
  const uint64 kIEpsilons
  const uint64 kNoIEpsilons
  const uint64 kOEpsilons
  const uint64 kNoOEpsilons
  const uint64 kILabelSorted
  const uint64 kNotILabelSorted
  const uint64 kOLabelSorted
  const uint64 kNotOLabelSorted
  const uint64 kWeighted
  const uint64 kUnweighted
  const uint64 kCyclic
  const uint64 kAcyclic
  const uint64 kInitialCyclic
  const uint64 kInitialAcyclic
  const uint64 kTopSorted
  const uint64 kNotTopSorted
  const uint64 kAccessible
  const uint64 kNotAccessible
  const uint64 kCoAccessible
  const uint64 kNotCoAccessible
  const uint64 kString
  const uint64 kNotString
  const uint64 kWeightedCycles
  const uint64 kUnweightedCycles
  const uint64 kNullProperties
  const uint64 kCopyProperties
  const uint64 kIntrinsicProperties
  const uint64 kExtrinsicProperties
  const uint64 kSetStartProperties
  const uint64 kSetFinalProperties
  const uint64 kAddStateProperties
  const uint64 kAddArcProperties
  const uint64 kSetArcProperties
  const uint64 kDeleteStatesProperties
  const uint64 kDeleteArcsProperties
  const uint64 kStateSortProperties
  const uint64 kArcSortProperties
  const uint64 kILabelInvariantProperties
  const uint64 kOLabelInvariantProperties
  const uint64 kWeightInvariantProperties
  const uint64 kAddSuperFinalProperties
  const uint64 kRmSuperFinalProperties
  const uint64 kBinaryProperties
  const uint64 kTrinaryProperties
  const uint64 kPosTrinaryProperties
  const uint64 kNegTrinaryProperties
  const uint64 kFstProperties

  # ArcIterator flags.
  const uint8 kArcILabelValue
  const uint8 kArcOLabelValue
  const uint8 kArcWeightValue
  const uint8 kArcNextStateValue
  const uint8 kArcNoCache
  const uint8 kArcValueFlags
  const uint8 kArcFlags

  # EncodeMapper flags.
  const uint8 kEncodeLabels
  const uint8 kEncodeWeights
  const uint8 kEncodeFlags

  # Default argument constants.
  const float kDelta
  const float kShortestDelta
  const int kNoLabel
  const int kNoStateId
  const int64 kNoSymbol

  enum ClosureType:
    CLOSURE_STAR
    CLOSURE_PLUS

  enum ComposeFilter:
    AUTO_FILTER
    NULL_FILTER
    SEQUENCE_FILTER
    ALT_SEQUENCE_FILTER
    MATCH_FILTER
    TRIVIAL_FILTER

  cdef cppclass ComposeOptions:

    ComposeOptions(bool, ComposeFilter)

  enum DeterminizeType:
    DETERMINIZE_FUNCTIONAL
    DETERMINIZE_NONFUNCTIONAL
    DETERMINIZE_DISAMBIGUATE

  enum EncodeType:
    DECODE
    ENCODE

  enum EpsNormalizeType:
    EPS_NORM_INPUT
    EPS_NORM_OUTPUT

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603
  ctypedef enum ProjectType:
    PROJECT_INPUT "fst::ProjectType::INPUT"
    PROJECT_OUTPUT "fst::ProjectType::OUTPUT"

  enum QueueType:
    TRIVIAL_QUEUE
    FIFO_QUEUE
    LIFO_QUEUE
    SHORTEST_FIRST_QUEUE
    TOP_ORDER_QUEUE
    STATE_ORDER_QUEUE
    SCC_QUEUE
    AUTO_QUEUE
    OTHER_QUEUE

  # This is a templated struct at the C++ level, but Cython does not support
  # templated structs unless we pretend they are full-blown classes.
  cdef cppclass RandGenOptions[RandArcSelection]:

    RandGenOptions(const RandArcSelection &, int32, int32, bool, bool)

  enum ReplaceLabelType:
    REPLACE_LABEL_NEITHER
    REPLACE_LABEL_INPUT
    REPLACE_LABEL_OUTPUT
    REPLACE_LABEL_BOTH

  enum ReweightType:
    REWEIGHT_TO_INITIAL
    REWEIGHT_TO_FINAL

  cdef cppclass SymbolTableTextOptions:

    SymbolTableTextOptions(bool)

  # This is actually a nested class, but Cython doesn't need to know that.
  cdef cppclass SymbolTableIterator "fst::SymbolTable::iterator":

      SymbolTableIterator(const SymbolTableIterator &)

      cppclass value_type:

        int64 Label()
        string Symbol()

      # When wrapped in a unique_ptr siter.Label() and siter.Symbol() are
      # ambiguous to Cython because there's no way to make the -> explicit.
      # This hacks around that.
      const value_type &Pair "operator*"()

      SymbolTableIterator &operator++()

      bool operator==(const SymbolTableIterator &, const SymbolTableIterator &)

      bool operator!=(const SymbolTableIterator &, const SymbolTableIterator &)


  # Symbol tables.
  cdef cppclass SymbolTable:

    @staticmethod
    int64 kNoSymbol

    SymbolTable()

    SymbolTable(const string &)

    @staticmethod
    SymbolTable *Read(const string &)

    # Aliased for overload.
    @staticmethod
    SymbolTable *ReadStream "Read"(istream &, const string &)

    @staticmethod
    SymbolTable *ReadText(const string &, const SymbolTableTextOptions &)

    int64 AddSymbol(const string &, int64)

    int64 AddSymbol(const string &)

    SymbolTable *Copy()

    # Aliased for overload.
    string FindSymbol "Find"(int64)

    # Aliased for overload.
    int64 FindIndex "Find"(const string &)

    # Aliased for overload.
    bool MemberSymbol "Member"(const string &)

    # Aliased for overload.
    bool MemberIndex "Member"(int64)

    void AddTable(const SymbolTable &)

    int64 GetNthKey(ssize_t)

    const string &Name()

    void SetName(const string &)

    const string &CheckSum()

    const string &LabeledCheckSum()

    bool Write(ostream &)

    bool Write(const string &)

    bool WriteText(ostream &)

    bool WriteText(const string &)

    SymbolTableIterator begin()

    SymbolTableIterator end()

    int64 AvailableKey()

    size_t NumSymbols()

  SymbolTable *CompactSymbolTable(const SymbolTable &syms)

  SymbolTable *MergeSymbolTable(const SymbolTable &, const SymbolTable &,
                                bool *)

  SymbolTable *FstReadSymbols(const string &, bool)

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603.
  ctypedef enum TokenType:
    SYMBOL "fst::TokenType::SYMBOL"
    BYTE "fst::TokenType::BYTE"
    UTF8 "fst::TokenType::UTF8"


cdef extern from "<fst/script/fstscript.h>" namespace "fst::script" nogil:

  cdef cppclass WeightClass:

    WeightClass()

    WeightClass(const WeightClass &)

    WeightClass(const string &, const string &)

    const string &Type()

    string ToString()

    bool Member()

    @staticmethod
    const WeightClass &Zero(const string &)

    @staticmethod
    const WeightClass &One(const string &)

    @staticmethod
    const WeightClass &NoWeight(const string &)

  # Alias.
  cdef bool Eq "operator=="(const WeightClass &, const WeightClass &)

  # Alias.
  cdef bool Ne "operator!="(const WeightClass &, const WeightClass &)

  cdef WeightClass Plus(const WeightClass &, const WeightClass &)

  cdef WeightClass Times(const WeightClass &, const WeightClass &)

  cdef WeightClass Divide(const WeightClass &, const WeightClass &)

  cdef WeightClass Power(const WeightClass &, size_t)

  cdef cppclass ArcClass:

    ArcClass(const ArcClass &)

    ArcClass(int64, int64, const WeightClass &, int64)

    int64 ilabel
    int64 olabel
    WeightClass weight
    int64 nextstate

  cdef cppclass FstClass:

    FstClass(const FstClass &)

    @staticmethod
    FstClass *Read(const string &)

    # Aliased for overload.
    @staticmethod
    FstClass *ReadStream "Read"(istream &, const string &)

    int64 Start()

    WeightClass Final(int64)

    size_t NumArcs(int64)

    size_t NumInputEpsilons(int64)

    size_t NumOutputEpsilons(int64)

    const string &ArcType()

    const string &FstType()

    const SymbolTable *InputSymbols()

    const SymbolTable *OutputSymbols()

    const string &WeightType()

    bool Write(const string &)

    bool Write(ostream &, const string &)

    uint64 Properties(uint64, bool)

    bool ValidStateId(int64)

  cdef cppclass MutableFstClass(FstClass):

    bool AddArc(int64, const ArcClass &)

    int64 AddState()

    void AddStates(size_t)

    bool DeleteArcs(int64, size_t)

    bool DeleteArcs(int64)

    bool DeleteStates(const vector[int64] &)

    void DeleteStates()

    SymbolTable *MutableInputSymbols()

    SymbolTable *MutableOutputSymbols()

    int64 NumStates()

    bool ReserveArcs(int64, size_t)

    void ReserveStates(int64)

    bool SetStart(int64)

    bool SetFinal(int64, const WeightClass &)

    void SetInputSymbols(const SymbolTable *)

    void SetOutputSymbols(const SymbolTable *)

    void SetProperties(uint64, uint64)

  cdef cppclass VectorFstClass(MutableFstClass):

   VectorFstClass(const FstClass &)

   VectorFstClass(const string &)

  cdef cppclass EncodeMapperClass:

    EncodeMapperClass(const string &, uint32, EncodeType)

    # Aliased to __call__ as Cython doesn't have good support for operator().
    ArcClass __call__ "operator()"(const ArcClass &)

    const string &ArcType()

    const string &WeightType()

    uint32 Flags()

    uint64 Properties(uint64)

    @staticmethod
    EncodeMapperClass *Read(const string &)

    # Aliased for overload.
    @staticmethod
    EncodeMapperClass *ReadStream "Read"(istream &, const string &)

    bool Write(const string &)

    # Aliased for overload.
    bool WriteStream "Write"(ostream &, const string &)

    const SymbolTable *InputSymbols()

    const SymbolTable *OutputSymbols()

    void SetInputSymbols(const SymbolTable *)

    void SetOutputSymbols(const SymbolTable *)

  cdef cppclass ArcIteratorClass:

    ArcIteratorClass(const FstClass &, int64)

    bool Done()

    ArcClass Value()

    void Next()

    void Reset()

    void Seek(size_t)

    size_t Position()

    uint8 Flags()

    void SetFlags(uint8, uint8)

  cdef cppclass MutableArcIteratorClass:

    MutableArcIteratorClass(MutableFstClass *, int64)

    bool Done()

    ArcClass Value()

    void Next()

    void Reset()

    void Seek(size_t)

    void SetValue(const ArcClass &)

    size_t Position()

    uint8 Flags()

    void SetFlags(uint8, uint8)

  cdef cppclass StateIteratorClass:

    StateIteratorClass(const FstClass &)

    bool Done()

    int64 Value()

    void Next()

    void Reset()


ctypedef pair[int64, const FstClass *] LabelFstClassPair

ctypedef pair[int64, int64] LabelPair


cdef extern from "<fst/script/fstscript.h>" namespace "fst::script" nogil:

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603
  ctypedef enum ArcFilterType:
    ANY_ARC_FILTER "fst::script::ArcFilterType::ANY"
    EPSILON_ARC_FILTER "fst::script::ArcFilterType::EPSILON"
    INPUT_EPSILON_ARC_FILTER "fst::script::ArcFilterType::INPUT_EPSILON"
    OUTPUT_EPSILON_ARC_FILTER "fst::script::ArcFilterType::OUTPUT_EPSILON"

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603
  ctypedef enum ArcSortType:
    ILABEL_SORT "fst::script::ArcSortType::ILABEL"
    OLABEL_SORT "fst::script::ArcSortType::OLABEL"

  cdef void ArcSort(MutableFstClass *, ArcSortType)

  cdef ClosureType GetClosureType(bool)

  cdef void Closure(MutableFstClass *, ClosureType)

  cdef unique_ptr[FstClass] CompileFstInternal(istream &,
                                               const string &,
                                               const string &,
                                               const string &,
                                               const SymbolTable *,
                                               const SymbolTable *,
                                               const SymbolTable*,
                                               bool,
                                               bool,
                                               bool,
                                               bool,
                                               bool)

  cdef void Compose(FstClass &, FstClass &, MutableFstClass *,
                    const ComposeOptions &)

  cdef void Concat(MutableFstClass *, const FstClass &)

  cdef void Connect(MutableFstClass *)

  cdef unique_ptr[FstClass] Convert(const FstClass &, const string &)

  cdef void Decode(MutableFstClass *, const EncodeMapperClass &)

  cdef cppclass DeterminizeOptions:

    DeterminizeOptions(float,
                       const WeightClass &,
                       int64,
                       int64,
                       DeterminizeType,
                       bool)

  cdef void Determinize(const FstClass &,
                        MutableFstClass *,
                        const DeterminizeOptions &)

  cdef cppclass DisambiguateOptions:

    DisambiguateOptions(float, const WeightClass &, int64, int64)

  cdef void Disambiguate(const FstClass &,
                         MutableFstClass *,
                         const DisambiguateOptions &)

  cdef void Difference(const FstClass &,
                       const FstClass &,
                       MutableFstClass *,
                       const ComposeOptions &)

  cdef void Draw(const FstClass &fst,
                 const SymbolTable *,
                 const SymbolTable *,
                 const SymbolTable *,
                 bool,
                 const string &,
                 float,
                 float,
                 bool,
                 bool,
                 float,
                 float,
                 int,
                 int,
                 const string &,
                 bool,
                 ostream &,
                 const string &)

  cdef void Encode(MutableFstClass *, EncodeMapperClass *)

  cdef EpsNormalizeType GetEpsNormalizeType(bool)

  cdef void EpsNormalize(const FstClass &, MutableFstClass *, EpsNormalizeType)

  cdef bool Equal(const FstClass &, const FstClass &, float)

  cdef bool Equivalent(const FstClass &, const FstClass &, float)

  cdef void Intersect(const FstClass &,
                      const FstClass &,
                      MutableFstClass *,
                      const ComposeOptions &)

  cdef void Invert(MutableFstClass *fst)

  cdef bool Isomorphic(const FstClass &, const FstClass &, float)

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603
  ctypedef enum MapType:
    ARC_SUM_MAPPER "fst::script::MapType::ARC_SUM"
    IDENTITY_MAPPER "fst::script::MapType::IDENTITY"
    INPUT_EPSILON_MAPPER "fst::script::MapType::INPUT_EPSILON"
    INVERT_MAPPER "fst::script::MapType::INVERT"
    OUTPUT_EPSILON_MAPPER "fst::script::MapType::OUTPUT_EPSILON"
    PLUS_MAPPER "fst::script::MapType::PLUS"
    QUANTIZE_MAPPER "fst::script::MapType::QUANTIZE"
    RMWEIGHT_MAPPER "fst::script::MapType::RMWEIGHT"
    SUPERFINAL_MAPPER "fst::script::MapType::SUPERFINAL"
    TIMES_MAPPER "fst::script::MapType::TIMES"
    TO_LOG_MAPPER "fst::script::MapType::TO_LOG"
    TO_LOG64_MAPPER "fst::script::MapType::TO_LOG64"
    TO_STD_MAPPER "fst::script::MapType::TO_STD"

  cdef unique_ptr[FstClass] Map(const FstClass &,
                                MapType,
                                float,
                                double,
                                const WeightClass &)

  cdef void Minimize(MutableFstClass *, MutableFstClass *, float, bool)

  cdef bool GetProjectType(const string &, ProjectType *)

  cdef void Project(MutableFstClass *, ProjectType)

  cdef void Print(const FstClass &,
                  ostream &,
                  const string &,
                  const SymbolTable *,
                  const SymbolTable *,
                  const SymbolTable *,
                  bool,
                  bool,
                  const string &)

  cdef void Prune(const FstClass &,
                  MutableFstClass *,
                  const WeightClass &,
                  int64, float)

  cdef void Prune(MutableFstClass *, const WeightClass &, int64, float)

  cdef void Push(const FstClass &,
                 MutableFstClass *,
                 uint8 flags,
                 ReweightType, float)

  cdef void Push(MutableFstClass *, ReweightType, float, bool)

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603
  ctypedef enum RandArcSelection:
    UNIFORM_ARC_SELECTOR "fst::script::RandArcSelection::UNIFORM"
    LOG_PROB_ARC_SELECTOR "fst::script::RandArcSelection::LOG_PROB"
    FAST_LOG_PROB_ARC_SELECTOR "fst::script::RandArcSelection::FAST_LOG_PROB"

  cdef bool RandEquivalent(const FstClass &,
                           const FstClass &,
                           int32,
                           const RandGenOptions[RandArcSelection] &,
                           float,
                           uint64)

  cdef void RandGen(const FstClass &,
                    MutableFstClass *,
                    const RandGenOptions[RandArcSelection] &,
                    uint64)

  cdef void Relabel(MutableFstClass *,
                    const SymbolTable *,
                    const SymbolTable *,
                    const string &, bool,
                    const SymbolTable *,
                    const SymbolTable *,
                    const string &,
                    bool)

  cdef void Relabel(MutableFstClass *,
                    const vector[LabelPair] &,
                    const vector[LabelPair] &)

  cdef cppclass ReplaceOptions:

     ReplaceOptions(int64, ReplaceLabelType, ReplaceLabelType, int64)

  cdef void Replace(const vector[LabelFstClassPair] &,
                    MutableFstClass *,
                    const ReplaceOptions &)

  cdef void Reverse(const FstClass &, MutableFstClass *, bool)

  cdef void Reweight(MutableFstClass *,
                     const vector[WeightClass] &,
                     ReweightType)

  cdef cppclass RmEpsilonOptions:

    RmEpsilonOptions(QueueType, bool, const WeightClass &, int64, float)

  cdef void RmEpsilon(MutableFstClass *, const RmEpsilonOptions &)

  cdef cppclass ShortestDistanceOptions:

    ShortestDistanceOptions(QueueType, ArcFilterType, int64, float)

  cdef void ShortestDistance(const FstClass &,
                             vector[WeightClass] *,
                             const ShortestDistanceOptions &)

  cdef void ShortestDistance(const FstClass &,
                             vector[WeightClass] *, bool,
                             float)

  cdef cppclass ShortestPathOptions:

    ShortestPathOptions(QueueType,
                        int32,
                        bool,
                        float,
                        const WeightClass &,
                        int64)

  cdef void ShortestPath(const FstClass &,
                         MutableFstClass *,
                         const ShortestPathOptions &)

  cdef void Synchronize(const FstClass &, MutableFstClass *)

  cdef bool TopSort(MutableFstClass *)

  cdef void Union(MutableFstClass *, const vector[FstClass *] &)

  cdef bool Verify(const FstClass &)


cdef extern from "<fst/script/getters.h>" namespace "fst::script" nogil:

  cdef bool GetArcSortType(const string &, ArcSortType *)

  cdef bool GetComposeFilter(const string &, ComposeFilter *)

  cdef bool GetDeterminizeType(const string &, DeterminizeType *)

  cdef uint8 GetEncodeFlags(bool, bool)

  cdef bool GetMapType(const string &, MapType *)

  cdef uint8 GetPushFlags(bool, bool, bool, bool)

  cdef bool GetQueueType(const string &, QueueType *)

  cdef bool GetRandArcSelection(const string &, RandArcSelection *)

  cdef bool GetReplaceLabelType(string, bool, ReplaceLabelType *)

  cdef ReweightType GetReweightType(bool)

  cdef bool GetTokenType(const string &, TokenType *)


cdef extern from "<fst/extensions/far/far.h>" namespace "fst" nogil:

  # TODO(wolfsonkin): Don't do this hack if Cython gets proper enum class
  # support: https://github.com/cython/cython/issues/1603
  ctypedef enum FarType:
    FAR_DEFAULT "fst::FarType::DEFAULT"
    FAR_STTABLE "fst::FarType::STTABLE"
    FAR_STLIST "fst::FarType::STLIST"
    FAR_FST "fst::FarType::FST"
    FAR_SSTABLE "fst::FarType::SSTABLE"

cdef extern from "<fst/extensions/far/getters.h>" \
    namespace "fst" nogil:

  string GetFarTypeString(FarType)


cdef extern from "<fst/extensions/far/getters.h>" \
    namespace "fst::script" nogil:

  bool GetFarType(const string &, FarType *)


cdef extern from "<fst/extensions/far/far-class.h>" \
    namespace "fst::script" nogil:

  cdef cppclass FarReaderClass:

    const string &ArcType()

    bool Done()

    bool Error()

    bool Find(const string &)

    const FstClass *GetFstClass()

    const string &GetKey()

    void Next()

    void Reset()

    FarType Type()

    # For simplicity, we always use the multiple-file one.
    @staticmethod
    unique_ptr[FarReaderClass] Open(const vector[string] &)

  cdef cppclass FarWriterClass:

    bool Add(const string &, const FstClass &)

    bool Error()

    const string &ArcType()

    FarType Type()

    @staticmethod
    unique_ptr[FarWriterClass] Create(const string &, const string &, FarType)



import sys
from pathlib import Path

from Cython.Build import cythonize
from setuptools import Extension, setup

sys.path.append('fstlib/cext')

setup(
    name='medicc2',
    version='0.8.1',
    author='Tom L Kaufmann, Marina Petkovic, Roland F Schwarz',
    author_email='tkau93@gmail.com, marina.55kovic@gmail.com, roland.f.schwarz@gmail.com',
    description='Whole-genome doubling-aware copy number phylogenies for cancer evolution',
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/schwarzlab/medicc2',
    classifiers=[
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
    ],
    packages=['medicc', 'fstlib', 'fstlib.cext'],
    scripts=['medicc2'],
    license='GPL-3',
    install_requires=[
        'numpy>=1.20.1',
        'pyyaml>=5.4.1',
        'pandas>=1.2.2',
        'joblib>=1.0.1',
        'biopython>=1.78',
        'scipy>=1.7',
        'matplotlib>=3.3.4'
    ],
    package_data={
        "medicc": ["objects/*.fst", "objects/*.bed", "logging_conf.yaml"],
        "fstlib": ["logging_conf.yaml", "cext/*.pxd", "cext/*.pyx", "cext/*.h"],
    },
    ext_modules = cythonize([
        Extension("fstlib.cext.pywrapfst", 
                  ["fstlib/cext/pywrapfst.pyx"],
                  include_dirs=['fstlib/cext'],
                  libraries=["fst", "fstfar", "fstscript", "fstfarscript"],
                  extra_compile_args=['-std=c++17'],
                  language = "c++"),

        Extension("fstlib.cext.ops", 
                  ["fstlib/cext/ops.pyx"],
                  include_dirs=['fstlib/cext'],
                  libraries=["fst", "fstfar", "fstscript", "fstfarscript"],
                  extra_compile_args=['-std=c++17'],
                  language = "c++")
    ])
)



import logging

import fstlib
import numpy as np

from medicc import tools

SYMBOL_GAP = 'gap'

# prepare logger
logger = logging.getLogger(__name__)


def create_symbol_table(max_cn=8, separator='X'):
    alphabet = [tools.int2hex(x) for x in range(max_cn+1)]
    symbol_table = fstlib.SymbolTable()
    symbol_table.add_symbol(SYMBOL_GAP)
    for s in alphabet:
        symbol_table.add_symbol(s)

    if separator is not None and separator != '':
        symbol_table.add_symbol(separator)
    return symbol_table


def _get_int_cns_from_symbol_table(symbol_table, separator='X'):
    cns = {x[1]: tools.hex2int(x[1]) for x in symbol_table if x[0] != 0 and x[1] != separator}
    return cns


def create_1step_amp_fst(symbol_table, separator='X'):

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)

    myfst.add_states(2)
    myfst.set_start(0)
    myfst.set_final(0, 0)
    myfst.set_final(1, 0)

    myfst.add_arcs(0, [(s, s, 0, 0) for s in cns.keys()])
    if separator is not None and separator != '':
        myfst.add_arc(0, (separator, separator, 0, 0))  # add separator
    myfst.add_arcs(0, [(s, t, 1, 1) for s in cns.keys() for t in cns.keys()
                       if ((cns[t]-cns[s]) == 1) & (s != '0')])  # transitions from 0->1
    myfst.add_arcs(1, [(s, t, 0, 1) for s in cns.keys() for t in cns.keys() if (
        (cns[t]-cns[s]) == 1) & (s != '0')])  # extension of an open window
    if '0' in cns.keys():
        myfst.add_arc(1, ('0', '0', 0, 1))
    myfst.add_arcs(1, [(s, s, 0, 0) for s in cns.keys() if s != '0'])  # return
    if separator is not None and separator != '':
        myfst.add_arc(1, (separator, separator, 0, 0))

    return myfst


def create_1step_del_fst(symbol_table, separator='X', exclude_zero=False):

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)

    myfst.add_states(2)
    myfst.set_start(0)
    myfst.set_final(0, 0)
    myfst.set_final(1, 0)

    myfst.add_arcs(0, [(s, s, 0, 0) for s in cns.keys()])
    if separator is not None and separator != '':
        myfst.add_arc(0, (separator, separator, 0, 0))  # add separator
    if exclude_zero:
        myfst.add_arcs(0, [(s, t, 1, 1) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == 1 and t != '0'])  # 0->1
        myfst.add_arcs(1, [(s, t, 0, 1) for s in cns.keys() for t in cns.keys()
                           if (cns[s]-cns[t]) == 1 and t != '0'])  # extend an open window
    else:
        myfst.add_arcs(0, [(s, t, 1, 1) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == 1])  # 0->1
        myfst.add_arcs(1, [(s, t, 0, 1) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == 1])  # extend an open window
    if '0' in cns.keys():
        myfst.add_arc(1, ('0', '0', 0, 1))
    myfst.add_arcs(1, [(s, s, 0, 0) for s in cns.keys() if s != '0'])  # return
    if separator is not None and separator != '':
        myfst.add_arc(1, (separator, separator, 0, 0))

    return myfst


def create_loh_fst(symbol_table, separator='X'):

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)

    myfst.add_states(9)
    myfst.set_start(0)
    for i in range(9):
        myfst.set_final(i, 0)

    ## match 0 -> 0
    myfst.add_arcs(0, [(s, s, 0, 0) for s in cns.keys()])
    if separator is not None and separator != '':
        myfst.add_arc(0, (separator, separator, 0, 0))  # add separator

    ## others
    for state in range(1, 9):
        cost = state
        myfst.add_arcs(0, [(s, t, cost, state) for s in cns.keys()
                           for t in cns.keys() if (cns[s]-cns[t]) == cost and t == '0'])
        myfst.add_arcs(state, [(s, t, 0, state) for s in cns.keys()
                               for t in cns.keys() if (cns[s]-cns[t]) <= cost and t == '0'])
        myfst.add_arcs(state, [(s, t, cns[s]-cns[t]-cost, cns[s]-cns[t]) for s in cns.keys()
                               for t in cns.keys() if (cns[s]-cns[t]) > cost and t == '0'])
        myfst.add_arcs(state, [(s, s, 0, 0) for s in cns.keys()])
        if separator is not None and separator != '':
            myfst.add_arc(state, (separator, separator, 0, 0))

    return myfst


def create_1step_WGD_fst(symbol_table, separator='X', wgd_cost=1, minimize=True, wgd_x2=False,
                         total_cn=False):

    if total_cn:
        wgd_distance = 2.
    else:
        wgd_distance = 1.

    cns = _get_int_cns_from_symbol_table(symbol_table, separator)

    W = fstlib.Fst()
    W.set_input_symbols(symbol_table)
    W.set_output_symbols(symbol_table)
    W.add_states(3)
    W.set_start(0)
    W.set_final(0, 0)
    W.set_final(1, 0)
    W.set_final(2, 0)
    if wgd_x2:
        W.add_arcs(0, [(s, t, wgd_cost, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]/cns[s]) == 2.)])
    else:
        W.add_arcs(0, [(s, t, wgd_cost, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]-cns[s]) == wgd_distance)])

    W.add_arc(1, ('0', '0', 0, 1))
    if wgd_x2:
        W.add_arcs(1, [(s, t, 0, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]/cns[s]) == 2.)])
    else:
        W.add_arcs(1, [(s, t, 0, 1) for s in cns.keys()
                    for t in cns.keys() if (s != '0') and ((cns[t]-cns[s]) == wgd_distance)])
    W.add_arc(0, ('0', '0', 0, 0))
    if separator is not None and separator != '':
        W.add_arc(0, (separator, separator, 0, 0))
        W.add_arc(1, (separator, separator, 0, 1))
    W.add_arc(1, ('0', '0', 0, 1))

    W.add_arcs(0, [(s, s, 0, 2) for s in cns.keys() if s != '0'])
    W.add_arcs(2, [(s, s, 0, 2) for s in cns.keys()])
    W.add_arc(2, ('0', '0', 0, 2))
    if separator is not None and separator != '':
        W.add_arc(2, (separator, separator, 0, 2))
    W.arcsort('olabel')
    if minimize:
        W = fstlib.encode_determinize_minimize(W)

    return W


def create_filter_fst(allowed_symbols, symbol_table, separator='X'):
    F = fstlib.Fst()
    F.set_input_symbols(symbol_table)
    F.set_output_symbols(symbol_table)
    F.add_states(1)
    F.set_start(0)
    F.set_final(0, 0)
    F.add_arcs(0, [(s, s, 0, 0) for s in allowed_symbols])
    F.add_arc(0, (separator, separator, 0, 0))

    return F


def create_nstep_fst(n, one_step_fst, minimize=True):
    # Extend 1step FST
    nstep_fst = one_step_fst
    nstep_fst.arcsort(sort_type='olabel')

    for _ in range(n):
        nstep_fst = fstlib.compose(nstep_fst, one_step_fst)
        if minimize:
            nstep_fst = fstlib.encode_determinize_minimize(nstep_fst)
        nstep_fst.arcsort(sort_type='olabel')

    return nstep_fst


def create_copynumber_fst(symbol_table, sep='X', enable_wgd=False, wgd_cost=1, max_num_wgds=3,
                          wgd_x2=False, total_cn=False):
    """ Creates the tree FST T which computes the asymmetric MED. """
    n = len(_get_int_cns_from_symbol_table(symbol_table, sep))
    X1step = create_1step_del_fst(symbol_table, sep, exclude_zero=True)
    X = create_nstep_fst(n-1, X1step)
    XX = fstlib.encode_determinize_minimize(X*~X)
    LOH = create_loh_fst(symbol_table, sep)

    if enable_wgd:
        W1step = create_1step_WGD_fst(symbol_table, sep, wgd_cost=wgd_cost,
                                      minimize=False)
        n = int(min(max_num_wgds, np.floor(np.log2(n))))
        if n > 1:
            W = create_nstep_fst(n-1, W1step)
        else:
            W = W1step
        T = LOH * W * XX
    else:
        T = LOH * XX

    return T


def create_phasing_fsa_from_strings(allele_a, allele_b, symbol_table, sep='X'):
    myfst = fstlib.Fst(arc_type='standard')
    myfst.set_input_symbols(symbol_table)
    myfst.set_output_symbols(symbol_table)
    prev_state = myfst.add_state()
    myfst.set_start(0)
    for a, b in zip(allele_a, allele_b):
        curr_state = myfst.add_state()
        myfst.add_arc(prev_state, (a, a, 0, curr_state))
        myfst.add_arc(prev_state, (b, b, 0, curr_state))
        prev_state = curr_state

    myfst.set_final(prev_state, 0)
    return myfst



import numpy as np
from scipy import stats


def _remove_duplicates(D):
    """ removes duplicates (entries with D=0) starting from the highest index. """
    finished=False
    while not finished:
        triu = np.triu_indices(D.shape[0], k=1)
        duplicates = np.flatnonzero(D[triu]==0)
        if len(duplicates)>0:
            dupe = duplicates[0]
            sel = max(triu[0][dupe], triu[1][dupe])
            D = np.delete(D, sel, axis=0)
            D = np.delete(D, sel, axis=1)
        else:
            finished=True
    return D

def star_topology_test(D):
    if D.shape[0] == 2:
        return 1
    """ Tests the distance matrix D for star topology. """
    ## check if any non-diagonal entry is zero
    D = _remove_duplicates(D)

    ## extract lower triangle matrix
    p=D[np.triu_indices(D.shape[0], k=1)]

    ## set up |p| x |samples| matrix T of zeros and ones mapping branch lengths
    ## in a star topology setting to the pairwise distances (e.g. dist(a,b) = branch_a + branch_b)
    count=0
    T = np.zeros((len(p), D.shape[1]))
    for i in range(T.shape[1]-1):
        for j in range(i+1, T.shape[1]):
            T[count,[i,j]]=1
            count = count + 1

    ## now find the optimal branch lengths b for our p
    Tinv = np.linalg.inv(np.matmul(T.T, T))
    bopt = np.matmul(np.matmul(p.T, T), Tinv).T
 
    ## scale the residuals by the square root of their variance (their sd)
    ## the estimates p are normal distributed with mean t (time of divergence) and variance t.
    res = (np.matmul(T, bopt)-p) / np.sqrt(p)
    res = np.matmul(res.T, res)

    ## the sum of squared residuals is therefore chisquare distributed with |p|-|samples| degrees
    ## of freedom. The "-|samples|" is because we optimize over the branch lengths (Tim Massingam,
    ## personal communication)
    
    prob = stats.chi2.cdf(res, len(p)-D.shape[1])
    retval = 1-prob
    return retval ## the p-value

def molecular_clock_test(D, normal_index=0):
    ## distances from diploid to all others but diploid
    p=np.delete(D, normal_index, axis=1)[normal_index, :] 
    ## remove other zeros
    p = p[p!=0]

    ## scale the residuals by the square root of their variance (their sd)
    ## the estimates p are normal distributed with mean t (time of divergence) and variance t.
    res = (np.mean(p) - p) / np.sqrt(p)
    res = np.matmul(res.T, res)
    
    ## the sum of squared residuals is therefore chisquare distributed with |p| degrees
    ## of freedom. (Tim Massingam, personal communication)

    prob = stats.chi2.cdf(res, len(p))
    return 1-prob ## the p-value


def MEDICCStatsError(Exception):
    pass
